import attr
import lxml.etree as etree


@attr.s
class BinningVariable(object):
    name = attr.ib()
    expression = attr.ib()
    valid_expression = attr.ib()

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "BinningVariable({})".format(
            ", ".join(["{}='{}'".format(key, value) for key, value in self.to_dict().items()])
        )

    def to_dict(self):
        return attr.asdict(self)

    def format_key(self, key):
        if key in ["expression"]:
            return "expr"
        elif key in ["valid_expression"]:
            return "validExpr"
        else:
            return key

    def to_xml(self):
        root = etree.Element("variable")

        for key, value in self.to_dict().items():
            elem = etree.SubElement(root, self.format_key(key))
            elem.text = value

        return root



import lxml.etree as etree

from .aggregator_avg import AggregatorAvg  # noqa: F401
from .aggregator_avg_outlier import AggregatorAvgOutlier  # noqa: F401
from .aggregator_minmax import AggregatorMinMax  # noqa: F401
from .aggregator_on_max_set import AggregatorOnMaxSet  # noqa: F401
from .aggregator_percentile import AggregatorPercentile  # noqa: F401
from .aggregator_sum import AggregatorSum  # noqa: F401


class Aggregators(object):
    def __init__(self, output_aggregators):
        self.output_aggregators = output_aggregators

    def to_xml(self):
        root = etree.Element("aggregators")

        for aggregator in self.output_aggregators:
            root.append(aggregator.to_xml())

        return root



import lxml.etree as etree

from .binning_variable import BinningVariable


class BinningVariables(object):
    def __init__(self, output_bands):
        if not isinstance(output_bands, list):
            raise ValueError("Provide a list with BinningVariable")

        self.output_bands = output_bands

        for band in self.output_bands:
            if not isinstance(band, BinningVariable):
                raise ValueError("Provide a list with BinningVariable")

    def to_xml(self):
        root = etree.Element("variables")

        for band in self.output_bands:
            if isinstance(band, BinningVariable):
                root.append(band.to_xml())
            else:
                raise ValueError("Provide a BinningVariable object")

        return root



import os
import subprocess
import tempfile

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization


def load_public_key(public_key_str):
    public_key = serialization.load_pem_public_key(
        bytes(public_key_str), backend=default_backend()
    )
    return public_key


def encrypt_data(data, public_key):
    with tempfile.NamedTemporaryFile(delete=False) as fp_pem, tempfile.NamedTemporaryFile(
        delete=False
    ) as fp_data:
        fp_pem.write(public_key)
        fp_pem.close()
        fp_data.write(data),
        fp_data.close()
        command = [
            "openssl",
            "pkeyutl",
            "-encrypt",
            "-inkey",
            fp_pem.name,
            "-pubin",
            "-in",
            fp_data.name,
        ]
        encrypt_data = subprocess.check_output(command)
        os.remove(fp_data.name)
        os.remove(fp_pem.name)
        return encrypt_data



"""HTTP functions. Inspired by EUMDAC request module."""

from typing import Any, Dict

import requests
from requests.adapters import HTTPAdapter, Retry

from .exceptions import HTTPError, QuotaExceededError
from .logging import logger


def get_adapter(max_retries: int, backoff_factor: float) -> HTTPAdapter:
    """
    Returns an HTTPAdapter able to handling retries.
    See: https://requests.readthedocs.io/en/latest/user/advanced/#example-automatic-retries

    :param max_retries: Max number of retries before failing
    :type max_retries: int
    :param backoff_factor: A backoff factor to apply between attempts after the second try
    :type backoff_factor: float
    """
    retries = Retry(
        total=max_retries,
        backoff_factor=backoff_factor,
        status_forcelist=[500, 502, 503, 504],
        allowed_methods=["HEAD", "GET", "OPTIONS", "POST", "PUT", "PATCH", "DELETE"],
    )
    return HTTPAdapter(max_retries=retries)


def request(
    method: str,
    url: str,
    max_retries: int = 5,
    backoff_factor: float = 0.25,
    **kwargs: Any,
) -> requests.Response:
    adapter = get_adapter(max_retries, backoff_factor)
    session = requests.Session()
    session.mount("https://", adapter)

    response = requests.Response()

    try:
        if hasattr(session, method):
            logger.debug(_pretty_print(method, url, kwargs))
            response = getattr(session, method.lower())(url, **kwargs)
            if response.status_code == 429:
                msg = response.json()["message"]
                raise QuotaExceededError(msg)
            response.raise_for_status()
        else:
            raise HTTPError(f"Operation not supported: {method}")
    except requests.exceptions.RetryError:
        raise HTTPError(f"Maximum retries ({max_retries}) reached for {method.capitalize()} {url}")

    return response


def get(url: str, **kwargs: Any) -> requests.Response:
    """Perform a GET HTTP request to the given `url` with the given parameters.

    Retries will be managed in a transparent way when making the request.

    :param url: URL to make the request to
    :type url: str

    :param kwargs: Extra arguments to pass to the request, refer to the requests library
        documentation for a list of possible arguments.
    :type kwargs: dict, optional

    :return: Response received from the server
    :rtype: requests.Response
    """
    return request("get", url, **kwargs)


def post(url: str, **kwargs: Any) -> requests.Response:
    """Perform a POST HTTP request to the given `url` with the given parameters.

    Retries and throttling will be managed in a transparent way when making the request.

    :param url: URL to make the request to
    :type url: str

    :param kwargs: Extra arguments to pass to the request, refer to the requests library
        documentation for a list of possible arguments.
    :type kwargs: dict, optional

    :return: Response received from the server
    :rtype: requests.Response
    """
    return request("post", url, **kwargs)


def patch(url: str, **kwargs: Any) -> requests.Response:
    """Perform a PATCH HTTP request to the given `url` with the given parameters.

    Retries and throttling will be managed in a transparent way when making the request.

    :param url: URL to make the request to
    :type url: str

    :param kwargs: Extra arguments to pass to the request, refer to the requests library
        documentation for a list of possible arguments.
    :type kwargs: dict, optional

    :return: Response received from the server
    :rtype: requests.Response
    """
    return request("patch", url, **kwargs)


def put(url: str, **kwargs: Any) -> requests.Response:
    """Perform a PUT HTTP request to the given `url` with the given parameters.

    Retries and throttling will be managed in a transparent way when making the request.

    :param url: URL to make the request to
    :type url: str

    :param kwargs: Extra arguments to pass to the request, refer to the requests library
        documentation for a list of possible arguments.
    :type kwargs: dict, optional

    :return: Response received from the server
    :rtype: requests.Response
    """
    return request("put", url, **kwargs)


def delete(url: str, **kwargs: Any) -> requests.Response:
    """Perform a DELETE HTTP request to the given `url` with the given parameters.

    Retries and throttling will be managed in a transparent way when making the request.

    :param url: URL to make the request to
    :type url: str

    :param kwargs: Extra arguments to pass to the request, refer to the requests library
        documentation for a list of possible arguments.
    :type kwargs: dict, optional

    :return: Response received from the server
    :rtype: requests.Response
    """
    return request("delete", url, **kwargs)


def _pretty_print(method: str, url: str, kwargs: Dict[str, Any]) -> str:
    """Returns a readable str of the given request."""
    pargs = {}
    for key in kwargs.keys():
        if key == "headers":
            headers = {
                header: kwargs[key][header]
                for key, header in kwargs.items()
                if header not in ["referer", "User-Agent"]
            }
            if len(headers) > 0:
                pargs[key] = headers
        elif key == "auth":
            if hasattr(kwargs[key], "token"):
                pargs[key] = f"Bearer {str(kwargs[key].token)}"  # type: ignore
            else:
                pargs[key] = f"{type(kwargs[key]).__name__}"  # type: ignore
        else:
            pargs[key] = kwargs[key]
    return f"Request: {method.upper()} {url}, payload: {pargs}"



from xml.sax.saxutils import escape

import attr
import lxml.etree as etree


@attr.s
class TargetBand(object):
    name = attr.ib()
    expression = attr.ib()
    type = attr.ib("float32")
    description = attr.ib(default=None)
    unit = attr.ib(default=None)
    no_data_value = attr.ib(default="NaN")

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "TargetBand({})".format(
            ", ".join(["{}='{}'".format(key, value) for key, value in self.to_dict().items()])
        )

    def to_dict(self):
        return attr.asdict(self)

    def to_xml(self):
        root = etree.Element("targetBand")

        for key, value in self.to_dict().items():
            elem = etree.SubElement(root, key)
            if key == "expression":
                elem.text = escape(value)
            else:
                elem.text = value

        return root



"""Code in this package is largely based on https://github.com/snap-contrib/snapista/.

Snapista requires a local instance of Snap.
We unwired the dependency as we only need to build or load the graph.
All the operators are statically listed.
"""

from .graph import Graph  # noqa: F401
from .operator import Operator  # noqa: F401
from .operatorparams import OperatorParams  # noqa: F401
from .target_band import TargetBand  # noqa: F401
from .target_band_descriptors import TargetBandDescriptors  # noqa: F401



import lxml.etree as etree


class TargetBandDescriptors(object):
    def __init__(self, target_bands):
        self.target_bands = target_bands

    def to_xml(self):
        root = etree.Element("targetBands")

        for target_band in self.target_bands:
            root.append(target_band.to_xml())

        return root



"""Data Tailor classes.

Most of the code has been taken from the official EUMDAC library
(https://pypi.org/project/eumdac/).
Changes have been made to the Chain class to create instances from a yaml file and
to encode it to base64 representation as required by the API.
"""
import sys
from base64 import b64encode
from collections.abc import Mapping
from dataclasses import asdict, dataclass
from typing import Any, Optional, Union

import yaml

from eocanvas.exceptions import InvalidChainError

if sys.version_info < (3, 9):
    from typing import MutableMapping
else:
    from collections.abc import MutableMapping


def _none_filter(*args: Any, **kwargs: Any) -> MutableMapping[str, Any]:
    """Build a mapping of '*args' and '**kwargs' removing None values."""
    return {k: v for k, v in dict(*args, **kwargs).items() if v is not None}


class AsDictMixin:
    """Base class adding an 'asdict' method that removes None values."""

    def asdict(self) -> MutableMapping[str, Any]:
        """Return the fields of the instance as a new dictionary mapping field names to
        field values, removing None values."""
        return asdict(self, dict_factory=_none_filter)  # type: ignore


@dataclass
class Filter(AsDictMixin):
    """Layer filter, a list of `bands` or layers for a given `product`.

    Attributes
    ----------
    - `id`: *str*
    - `name`: *str*
        Human readable name.
    - `product`: *str*
        Product that the filter applies to.
    - `bands`: *list[dict]*
        List of bands part of the filter, as dicts of {id, number, name}.
    """

    id: Optional[str] = None
    bands: Optional[list] = None  # type: ignore[type-arg]
    name: Optional[str] = None
    product: Optional[str] = None


@dataclass
class RegionOfInterest(AsDictMixin):
    """Region of interest, a geographical area defined by its `NSWE` coordinates.

    Attributes
    ----------
    - `id`: *str*
    - `name`: *str*
        Human readable name.
    - `description`: *str*
        Human readable description.
    - `NSWE`:
        North, south, west, east coordinates, in decimal degrees.
    """

    id: Optional[str] = None
    name: Optional[str] = None
    NSWE: Optional[str] = None
    description: Optional[str] = None


@dataclass
class Quicklook(AsDictMixin):
    """Configuration for generating quicklooks."""

    id: Optional[str] = None
    name: Optional[str] = None
    resample_method: Optional[str] = None
    stretch_method: Optional[str] = None
    product: Optional[str] = None
    format: Optional[str] = None
    nodatacolor: Optional[str] = None
    filter: Union[None, dict, Filter] = None  # type: ignore[type-arg]
    x_size: Optional[int] = None

    def __post_init__(self) -> None:
        """Prepare `filter` as a Filter instance if given as dict."""
        if self.filter is not None and isinstance(self.filter, dict):
            self.filter = Filter(**self.filter)


@dataclass
class Chain(AsDictMixin):
    """Chain configuration for Data Tailor customisation jobs."""

    __submodels = {"filter": Filter, "roi": RegionOfInterest, "quicklook": Quicklook}
    id: Optional[str] = None
    product: Optional[str] = None
    format: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None
    aggregation: Optional[str] = None
    projection: Optional[str] = None
    roi: Union[None, dict, RegionOfInterest] = None  # type: ignore[type-arg]
    filter: Union[None, dict, Filter] = None  # type: ignore[type-arg]
    quicklook: Union[None, dict, Quicklook] = None  # type: ignore[type-arg]
    resample_method: Optional[str] = None
    resample_resolution: Optional[list] = None  # type: ignore[type-arg]
    compression: Optional[dict] = None  # type: ignore[type-arg]
    xrit_segments: Optional[list] = None  # type: ignore[type-arg]

    def __post_init__(self) -> None:
        """Prepare attributes as an instance of their class if given as dict."""
        for name, model in self.__submodels.items():
            attr = getattr(self, name)
            if attr is not None and isinstance(attr, Mapping):
                setattr(self, name, model(**attr))

    @classmethod
    def from_file(cls, filepath: str) -> "Chain":
        with open(filepath) as f:
            try:
                return cls(**yaml.safe_load(f))
            except TypeError as exc:
                raise InvalidChainError(f"Invalid chain file: {exc}")

    def b64encode(self) -> str:
        return b64encode(yaml.dump(self.asdict()).encode()).decode()



from types import SimpleNamespace

from eocanvas.exceptions import SnapOperatorNotFound

from .operatorparams import OperatorParams
from .snap import Snap


class Operator(SimpleNamespace):
    def __init__(self, operator, **kwargs):
        self._snap_operators = Snap().operators
        if operator not in self._snap_operators:
            raise SnapOperatorNotFound(f"Operator {operator} not recognized.")
        self.operator = operator
        self._params = {**OperatorParams(self.operator).params, **kwargs}
        super().__init__(**self._params)

    def __str__(self):
        return "{}:\n\t{}".format(
            self.operator,
            "\n\t".join(["{}='{}'".format(key, value) for key, value in self.to_dict().items()]),
        )

    def __repr__(self):
        return "Operator('{}', {})".format(
            self.operator,
            ", ".join(["{}='{}'".format(key, value) for key, value in self.to_dict().items()]),
        )

    def to_dict(self):
        return dict([(name, getattr(self, name)) for name in list(self._params.keys())])

    def describe(self):
        """This function prints the human readable information about a SNAP operator

        Args:
            operator: SNAP operator

        Returns
            The human readable information about the provided SNAP operator.

        Raises:
            None.
        """
        operator = self._snap_operators[self.operator]
        print(f"Operator name: {operator['alias']}\n")
        print(f"Description: {operator['description']}")
        print(f"Authors: {operator['authors']}\n")
        print(f"{operator['name']}")
        print(f"Version: {operator['version']}\n")
        print("Parameters:\n")

        for name, param in operator["params"].items():
            print(
                f"\t{name}: {param['description']}\n\t\tDefault Value: {param['default_values']}\n"
            )
            print(f"\t\tPossible values: {param['values_set']}\n")

    @staticmethod
    def _get_formats(method):
        """This function provides a human readable list of SNAP Read or Write operator formats.

        Args:
            None.

        Returns
            Human readable list of SNAP Write operator formats.

        Raises:
            None.
        """

        if method == "Read":
            return Snap().read_formats
        elif method == "Write":
            return Snap().write_formats
        else:
            raise ValueError



from __future__ import annotations

import base64
import json
import os
import time
from dataclasses import dataclass, fields
from datetime import datetime
from logging import INFO
from typing import Any, Dict, List, Optional, Union

from .auth import Credentials, HTTPOAuth2, OAuthToken
from .config import URLs
from .exceptions import JobFailed, NotDownloadableError
from .http import delete, get, post
from .keystore import encrypt_data
from .logging import logger
from .utils import Singleton


def transform_data(input_dict, mapping):
    """Remaps the dictionary keys according to the mapping."""
    return {mapping.get(k, k): v for k, v in input_dict.items()}


def filter_dict_for_dataclass(data_class, input_dict):
    """Discards dictionary keys that don't match with any dataclass field."""
    data_class_fields = {f.name for f in fields(data_class)}
    return {k: v for k, v in input_dict.items() if k in data_class_fields}


class API(metaclass=Singleton):
    """Models the Serverless Functions API endpoints.

    Attributes:
        urls: The :class:`eocanvas.config.URLs` object that maps all the API endpoints
        credentials: A :class:`eocanvas.auth.Credentials` object with username and password
    """

    def __init__(
        self,
        urls: Optional[URLs] = None,
        credentials: Optional[Credentials] = None,
        log_level: int = INFO,
    ):
        """"""
        logger.setLevel(log_level)

        if urls is None:
            urls = URLs()

        if credentials is None:
            credentials = Credentials.load()

        token = OAuthToken(url=urls.token_url, credentials=credentials)

        self.urls = urls
        self.auth = HTTPOAuth2(token)

    def get_public_key(self) -> str:
        url = self.urls.get("key_detail", key_id="cert/public")
        response = get(url, auth=self.auth)
        return response.content

    def get_key(self, key_id: str) -> Key:
        url = self.urls.get("key_detail", key_id=key_id)
        response = get(url, auth=self.auth)
        return self._build_key(response.json())

    def get_keys(self) -> List[Key]:
        url = self.urls.get("key_list")
        response = get(url, auth=self.auth)
        return [self._build_key(data) for data in response.json()]

    def create_key(self, key: Key) -> Key:
        url = self.urls.get("key_list")
        post(url, json=key.asdict(), auth=self.auth)
        # The API response is empty. We return the key itself.
        return key

    def delete_key(self, key_id: str) -> None:
        url = self.urls.get("key_detail", key_id=key_id)
        delete(url, auth=self.auth)

    def get_process(self, process_id: str) -> Process:
        """Gets the details of a process.

        Returns:
            A :class:`eocanvas.api.Process` instance.
        """
        url = self.urls.get("process_detail", process_id=process_id)
        response = get(url, auth=self.auth)
        return self._build_process(response.json())

    def get_processes(self) -> List[Process]:
        """Gets the list of available processes.
        There should be only SNAP and Datatailor functions.

        Returns:
            A list of :class:`eocanvas.api.Process` instances.
        """
        url = self.urls.get("process_list")
        response = get(url, auth=self.auth)
        return [self._build_process(data) for data in response.json()]

    def exec_process(self, process: Process) -> Job:
        """Submits a process to the API.

        Returns:
            A :class:`eocanvas.api.Job` instance.
        """
        inputs = process.prepare_inputs()
        url = self.urls.get("process_execution", process_id=process.process_id)
        response = post(url, json=inputs, auth=self.auth)
        return self._build_job(response.json())

    def get_job(self, job_id) -> Job:
        """Gets the details of a job.

        Returns:
            A :class:`eocanvas.api.Job` instance.
        """
        url = self.urls.get("job_detail", job_id=job_id)
        response = get(url, auth=self.auth)
        return self._build_job(response.json())

    def get_jobs(self) -> List[Job]:
        """Gets the list of user submitted jobs.

        Returns:
            A list of :class:`eocanvas.api.Job` instances.
        """
        url = self.urls.get("job_list")
        response = get(url, auth=self.auth)
        return [self._build_job(data) for data in response.json()["jobs"]]

    def get_job_logs(self, job: Union[Job, str]) -> List[LogEntry]:
        """Gets the log entries for a job.

        Args:
            job: Either a :class:`eocanvas.api.Job` instance or the `job_id`

        Returns:
            A list of :class:`eocanvas.api.LogEntry` instances.
        """
        if isinstance(job, Job):
            job_id = job.job_id
        else:
            job_id = job

        url = self.urls.get("job_logs", job_id=job_id)
        response = get(url, auth=self.auth)
        return [self._build_log_entry(data) for data in response.json()]

    def get_job_results(self, job: Union[Job, str]) -> List[Result]:
        """Gets the results for a job.

        Args:
            job: Either a :class:`eocanvas.api.Job` instance or the `job_id`

        Returns:
            A list of :class:`eocanvas.api.Result` instances.
        """
        if isinstance(job, Job):
            job_id = job.job_id
        else:
            job_id = job

        url = self.urls.get("job_results", job_id=job_id)
        response = get(url, auth=self.auth)
        results = response.json()

        # If the results are paginated, on the first page the last item is the link
        # to the next one. In that case, we follow it.
        next_page = results[-1]
        while next_page["title"] == "next-page":
            results.pop()
            response = get(self.urls.base_url + next_page["href"], auth=self.auth)
            paginated_results = response.json()
            next_page = paginated_results[-1]
            # From the second page on, the last item is the link to the previous page
            # and the second last the link to the next, if any
            if next_page["title"] == "prev-page":
                next_page = paginated_results[-2]
                results.extend(paginated_results[:-1])
            else:
                results.extend(paginated_results)

        return [self._build_result(data) for data in results]

    def download_result(self, result: Result, download_dir: Optional[str] = None):
        if download_dir is None:
            download_dir = "."

        os.makedirs(download_dir, exist_ok=True)
        response = get(result.full_url, auth=self.auth, stream=True)
        filename = result.title.split("/")[-1]
        download_path = os.path.join(download_dir, filename)
        logger.info(f"Downloading {download_path}")
        with open(download_path, "wb") as f:
            for chunk in response.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)

    def _build_key(self, data) -> Key:
        mapping = {
            "creationDate": "creation_date",
            "expirationDate": "expiration_date",
            "expireSeconds": "expire_seconds",
            "type": "type_",
        }
        data = filter_dict_for_dataclass(Key, transform_data(data, mapping))
        data["api"] = self
        return Key(**data)

    def _build_process(self, data) -> Process:
        mapping = {"processId": "process_id"}
        data = filter_dict_for_dataclass(Process, transform_data(data, mapping))
        data["api"] = self
        return Process(**data)

    def _build_job(self, data) -> Job:
        mapping = {"jobID": "job_id"}
        data = filter_dict_for_dataclass(Job, transform_data(data, mapping))
        data["api"] = self
        return Job(**data)

    def _build_result(self, data) -> Result:
        data = filter_dict_for_dataclass(Result, data)
        data["api"] = self
        return Result(**data)

    def _build_log_entry(self, data) -> LogEntry:
        data = filter_dict_for_dataclass(LogEntry, data)
        # Truncate the nanoseconds to microseconds for compatibility with datetime
        data["timestamp"] = data["timestamp"][:26] + data["timestamp"][-6:]
        data["timestamp"] = datetime.fromisoformat(data["timestamp"])
        return LogEntry(**data)


@dataclass
class Input:
    key: str
    url: str
    keystore: Optional[Union[str, Key]] = None

    def asdict(self) -> Dict:
        value = {self.key: self.url}
        if self.keystore is not None:
            if isinstance(self.keystore, str):
                value["schema"] = f"keystore://{self.keystore}"
            else:
                value["schema"] = f"keystore://{self.keystore.name}"

        return value


@dataclass
class ConfigOption:
    sub_path: str
    uncompress: Optional[bool] = None

    def asdict(self) -> Dict:
        if self.uncompress is None:
            return {"subPath": self.sub_path}

        return {"uncompress": self.uncompress, "subPath": self.sub_path}


@dataclass
class Config:
    key: str
    options: ConfigOption

    def asdict(self):
        return {self.key: self.options.asdict()}


@dataclass
class Job:
    """Represents an API job, that is a submission of a process."""

    api: API
    job_id: str
    status: Optional[str]
    started: Optional[str]
    created: Optional[str] = None
    updated: Optional[str] = None
    finished: Optional[str] = None

    def refresh_from_api(self):
        """Reloads the job attributes from the API."""
        job = self.api.get_job(self.job_id)
        self.status = job.status
        self.created = job.created
        self.updated = job.updated
        self.finished = job.finished

    @property
    def completed(self) -> bool:
        return self.status == "completed"

    @property
    def logs(self) -> List[LogEntry]:
        return self.api.get_job_logs(self)

    @property
    def results(self) -> List[Result]:
        return self.api.get_job_results(self)


@dataclass
class Result:
    api: API
    href: str
    title: str
    rel: str = "result"

    @property
    def full_url(self) -> str:
        return self.api.urls.get("download", result_href=self.href)

    def download(self, download_dir: Optional[str] = None):
        if self.title.startswith("keystore"):
            raise NotDownloadableError(
                "External reference to the result, not served by this service."
            )
        else:
            return self.api.download_result(self, download_dir)


@dataclass
class LogEntry:
    timestamp: datetime
    message: str


@dataclass
class Key:
    """A key object as returned by the API.
    Keys are used to configure external storages such as S3, EODATA and WEkEO Drive.
    """

    name: str
    type_: Optional[str] = None
    config: Optional[KeyConfig] = None
    description: Optional[str] = None
    creation_date: Optional[str] = None
    owner: Optional[str] = None
    public: Optional[bool] = False
    api: Optional[API] = None
    expiration_date: Optional[str] = None
    expire_seconds: Optional[int] = 0

    def __post_init__(self):
        if self.config is not None:
            if isinstance(self.config, S3KeyConfig):
                self.type_ = "S3"
            elif isinstance(self.config, WebDavKeyConfig):
                self.type_ = "WEBDAV"

        if self.type_ is not None and self.type_ not in ("S3", "WEBDAV"):
            raise ValueError("Type must be either 'S3' or 'WEBDAV'")

        if self.api is None:
            self.api = API()

    def asdict(self) -> Dict:
        data = {
            f"{self.name}": {
                "type": self.type_,
                "expire": self.expire_seconds or 3600,
                "public": self.public,
                "description": self.description or self.name,
                "data": self.config.encode(),
            }
        }
        return data

    def create(self) -> Key:
        if self.config is None:
            raise ValueError("Cannot create a key without config")

        return self.api.create_key(self)

    def delete(self):
        return self.api.delete_key(self.name)


@dataclass
class _APIParam:
    """
    A simple dataclass with an optional api parameter.
    This is used to avoid `non-default argument 'foo' follows default argument`
    """

    api: Optional[API] = None

    def __post_init__(self):
        super().__post_init__()
        if self.api is None:
            self.api = API()


@dataclass
class KeyConfig:
    """A configuration object that encrypt either an S3 or a WEBDAV key"""

    def asdict(self):
        return {f.name: getattr(self, f.name) for f in fields(self) if f.name != "api"}

    def encode(self) -> str:
        encrypted = encrypt_data(json.dumps(self.asdict()).encode(), self.api.get_public_key())
        encoded = base64.b64encode(encrypted).decode("utf-8")
        return encoded


@dataclass
class _S3KeyParams:
    secret_key: str
    access_key: str
    bucket: str
    endpoint: str
    region: str

    def __post_init__(self):
        if not self.endpoint.startswith("https://"):
            raise ValueError("Endpoint does not look like a URL")


@dataclass
class _WebDavKeyParams:
    endpoint: str
    username: str
    password: str

    def __post_init__(self):
        if not self.endpoint.startswith("https://"):
            raise ValueError("Endpoint does not look like a URL")


@dataclass
class S3KeyConfig(KeyConfig, _APIParam, _S3KeyParams):
    """A configuration for S3-like keys."""


@dataclass
class WebDavKeyConfig(KeyConfig, _APIParam, _WebDavKeyParams):
    """A configuration for WebDav-like keys."""


@dataclass
class Process:
    """A process defines either a SNAP or a Datatailor function."""

    api: Optional[API] = None
    process_id: Optional[str] = None
    version: Optional[str] = None
    title: Optional[str] = None
    description: Optional[str] = None
    inputs: Optional[Any] = None
    output: Optional[Union[str, Key]] = None

    def __post_init__(self):
        if self.api is None:
            self.api = API()

    def prepare_inputs(self) -> Dict:
        inputs = {
            "inputs": {},  # To be defined in subclasses
            "outputs": {},
            "response": "raw",
            "subscriber": None,
        }
        if self.output is not None:
            if isinstance(self.output, str):
                keystore = self.output
            else:
                keystore = self.output.name

            inputs["outputs"] = {"output": {"format": {"schema": f"keystore://{keystore}"}}}

        return inputs

    def submit(self) -> Job:
        return self.api.exec_process(self)

    def run(
        self, job: Optional[Job] = None, download_dir: Optional[str] = None, download: bool = True
    ):
        if job is None:
            job = self.submit()
        return JobRunner(job, download).run(download_dir)


class JobRunner:
    def __init__(self, job: Job, download: bool = True):
        self.job = job
        self.download = download

    def run(self, download_dir: Optional[str] = None):
        sleep = 10
        status = self.job.status
        while status != "successful":
            now = datetime.now().isoformat()
            logger.info(f"Job: {self.job.job_id} - Status: {status} at {now}")
            if status == "failed":
                raise JobFailed(
                    f"Job {self.job.job_id} failed. Try checking the logs for more info."
                )
            assert status in ["accepted", "running"]
            time.sleep(sleep)
            sleep = sleep * 1.1
            self.job.refresh_from_api()
            status = self.job.status

        if self.download:
            for result in self.job.results:
                try:
                    result.download(download_dir)
                except NotDownloadableError:
                    logger.info(result.title)



from eocanvas.exceptions import SnapOperatorNotFound

from .snap import Snap


class OperatorParams:
    def __init__(self, operator):
        self._snap_operators = Snap().operators
        if operator not in self._snap_operators:
            raise SnapOperatorNotFound(f"Operator {operator} not recognized.")

        self.operator = operator
        params = self._snap_operators[operator]["params"].keys()
        self.params = {
            param: self._snap_operators[operator]["params"][param].get("default_values")
            for param in params
        }



import os
import sys
from importlib import resources
from pathlib import Path
from typing import Optional

import yaml


def get_credentials_dir(create: bool = False) -> Path:
    path = Path(os.getenv("EOCANVAS_CONFIG_DIR", Path.home()))
    if create:
        path.mkdir(exist_ok=True)
    return path


def get_credentials_filepath(create: bool = False) -> Path:
    return get_credentials_dir(create) / ".hdarc"


class URLs:
    def __init__(self, urlfile: Optional[str] = None):
        if urlfile is None:
            if sys.version_info >= (3, 9):
                with resources.as_file(resources.files("eocanvas").joinpath("urls.yaml")) as path:
                    with open(path) as f:
                        data = yaml.safe_load(f)
            else:  # python < 3.9
                with resources.path("eocanvas", "urls.yaml") as path:
                    with open(path) as f:
                        data = yaml.safe_load(f)

        else:
            with open(urlfile) as f:
                data = yaml.safe_load(f)

        try:
            self.token_url = data["token_url"]
            self.base_url = data["base_url"]
            self.endpoints = data["endpoints"]
        except KeyError as err:
            raise KeyError(f"Malformed url file: {err}")

    def get(self, endpoint_name: str, **kwargs) -> str:
        endpoint = self.endpoints.get(endpoint_name)
        if endpoint is None:
            raise ValueError(f"Endpoint {endpoint_name} not found in configuration.")
        return self.base_url + endpoint.format(**kwargs)



import json
import os

from eocanvas.exceptions import MalformedSnapError
from eocanvas.utils import Singleton


class Snap(metaclass=Singleton):
    def __init__(self, filename=None):
        if filename is None:
            filename = os.path.join(os.path.dirname(os.path.abspath(__file__)), "operators.json")
        self.filename = filename
        self._operators = None

    @property
    def operators(self):
        if self._operators is None:
            try:
                with open(self.filename) as f:
                    self._operators = json.load(f)
            except FileNotFoundError:
                raise FileNotFoundError(f"SNAP operators file {self.filename} not found.")
            except json.JSONDecodeError as exc:
                raise MalformedSnapError(
                    f"SNAP operators file{self.filename} is not valid JSON: {exc}"
                )

        return self._operators

    @property
    def read_formats(self):
        return [
            "ACE",
            "ACE2",
            "ALOS PALSAR CEOS",
            "ALOS-2 CEOS",
            "ALOS-2 GeoTIFF",
            "AlosAV2Dimap",
            "AlosPRIDimap",
            "Aquarius",
            "ARC_INFO_BIN_GRID",
            "ATS_L1_S3",
            "ATSR",
            "AVNIR-2",
            "Basic CEOS",
            "BEAM-DIMAP",
            "Binned_data_product",
            "Capella",
            "Complex Binary",
            "CosmoSkymed",
            "CSV",
            "DEIMOSDimap",
            "DSCOVER-L2",
            "ENVI",
            "ENVISAT",
            "ERS CEOS",
            "FAST L5",
            "Gamma",
            "Gaofen3",
            "GDAL-BT-READER",
            "GDAL-ERS-READER",
            "GDAL-GS7BG-READER",
            "GDAL-GSBG-READER",
            "GDAL-GTiff-READER",
            "GDAL-GTX-READER",
            "GDAL-HFA-READER",
            "GDAL-ILWIS-READER",
            "GDAL-KRO-READER",
            "GDAL-MFF-READER",
            "GDAL-NITF-READER",
            "GDAL-PCIDSK-READER",
            "GDAL-RMF-READER",
            "GDAL-RST-READER",
            "GDAL-SAGA-READER",
            "GDAL-SGI-READER",
            "Generic Binary",
            "GeoTIFF-BigTIFF",
            "GeoTIFF",
            "GETASSE30",
            "GOCI-L1B",
            "GTOPO30",
            "HDF-EOS",
            "HGT",
            "HICO-L1B",
            "HICO",
            "IceyeProduct",
            "IkonosGeoTIFF",
            "IMAGE",
            "JERS CEOS",
            "JPEG2000",
            "Kompsat2GeoTIFF",
            "Kompsat5",
            "L3_ENV_MER",
            "Landsat8GeoTIFF15m",
            "Landsat8GeoTIFF30m",
            "LandsatC2L2GeoTIFF",
            "LandsatGeoTIFF",
            "LandsatL2GeoTIFF",
            "Matlab",
            "MER_L1_S3",
            "MER_L2_S3",
            "METOP_AVHRR_3_L1B",
            "MODIS-L1B",
            "MODIS",
            "MUSCATE",
            "NetCDF-BEAM",
            "NetCDF-CF",
            "NetCDF",
            "NOAA_AVHRR_3_L1B",
            "NOAA_POD_AVHRR_HRPT",
            "OCM2-L1B",
            "PACE-L1B",
            "PAZ",
            "PGX",
            "PleiadesDimap",
            "PolSARPro",
            "PRISM",
            "RADARSAT-1",
            "RADARSAT-2",
            "RapidEyeGeoTIFF",
            "RapidEyeNITF",
            "RAT",
            "RCM",
            "RISAT-1",
            "SAOCOM",
            "SeaDAS-ANC",
            "SeaDAS-Browse",
            "SeaDAS-L1",
            "SeaDAS-L2",
            "SeaDAS-L3",
            "SeaSat",
            "Sen3_SLSTRL1B_1km",
            "Sen3_SLSTRL1B_500m",
            "Sen3",
            "SENTINEL-1",
            "SENTINEL-2-MSI-10M",
            "SENTINEL-2-MSI-20M",
            "SENTINEL-2-MSI-60M",
            "SENTINEL-2-MSI-L1B-MultiRes",
            "SENTINEL-2-MSI-MultiRes-UTM01N",
            "SENTINEL-2-MSI-MultiRes-UTM01S",
            "SENTINEL-2-MSI-MultiRes-UTM02N",
            "SENTINEL-2-MSI-MultiRes-UTM02S",
            "SENTINEL-2-MSI-MultiRes-UTM03N",
            "SENTINEL-2-MSI-MultiRes-UTM03S",
            "SENTINEL-2-MSI-MultiRes-UTM04N",
            "SENTINEL-2-MSI-MultiRes-UTM04S",
            "SENTINEL-2-MSI-MultiRes-UTM05N",
            "SENTINEL-2-MSI-MultiRes-UTM05S",
            "SENTINEL-2-MSI-MultiRes-UTM06N",
            "SENTINEL-2-MSI-MultiRes-UTM06S",
            "SENTINEL-2-MSI-MultiRes-UTM07N",
            "SENTINEL-2-MSI-MultiRes-UTM07S",
            "SENTINEL-2-MSI-MultiRes-UTM08N",
            "SENTINEL-2-MSI-MultiRes-UTM08S",
            "SENTINEL-2-MSI-MultiRes-UTM09N",
            "SENTINEL-2-MSI-MultiRes-UTM09S",
            "SENTINEL-2-MSI-MultiRes-UTM10N",
            "SENTINEL-2-MSI-MultiRes-UTM10S",
            "SENTINEL-2-MSI-MultiRes-UTM11N",
            "SENTINEL-2-MSI-MultiRes-UTM11S",
            "SENTINEL-2-MSI-MultiRes-UTM12N",
            "SENTINEL-2-MSI-MultiRes-UTM12S",
            "SENTINEL-2-MSI-MultiRes-UTM13N",
            "SENTINEL-2-MSI-MultiRes-UTM13S",
            "SENTINEL-2-MSI-MultiRes-UTM14N",
            "SENTINEL-2-MSI-MultiRes-UTM14S",
            "SENTINEL-2-MSI-MultiRes-UTM15N",
            "SENTINEL-2-MSI-MultiRes-UTM15S",
            "SENTINEL-2-MSI-MultiRes-UTM16N",
            "SENTINEL-2-MSI-MultiRes-UTM16S",
            "SENTINEL-2-MSI-MultiRes-UTM17N",
            "SENTINEL-2-MSI-MultiRes-UTM17S",
            "SENTINEL-2-MSI-MultiRes-UTM18N",
            "SENTINEL-2-MSI-MultiRes-UTM18S",
            "SENTINEL-2-MSI-MultiRes-UTM19N",
            "SENTINEL-2-MSI-MultiRes-UTM19S",
            "SENTINEL-2-MSI-MultiRes-UTM20N",
            "SENTINEL-2-MSI-MultiRes-UTM20S",
            "SENTINEL-2-MSI-MultiRes-UTM21N",
            "SENTINEL-2-MSI-MultiRes-UTM21S",
            "SENTINEL-2-MSI-MultiRes-UTM22N",
            "SENTINEL-2-MSI-MultiRes-UTM22S",
            "SENTINEL-2-MSI-MultiRes-UTM23N",
            "SENTINEL-2-MSI-MultiRes-UTM23S",
            "SENTINEL-2-MSI-MultiRes-UTM24N",
            "SENTINEL-2-MSI-MultiRes-UTM24S",
            "SENTINEL-2-MSI-MultiRes-UTM25N",
            "SENTINEL-2-MSI-MultiRes-UTM25S",
            "SENTINEL-2-MSI-MultiRes-UTM26N",
            "SENTINEL-2-MSI-MultiRes-UTM26S",
            "SENTINEL-2-MSI-MultiRes-UTM27N",
            "SENTINEL-2-MSI-MultiRes-UTM27S",
            "SENTINEL-2-MSI-MultiRes-UTM28N",
            "SENTINEL-2-MSI-MultiRes-UTM28S",
            "SENTINEL-2-MSI-MultiRes-UTM29N",
            "SENTINEL-2-MSI-MultiRes-UTM29S",
            "SENTINEL-2-MSI-MultiRes-UTM30N",
            "SENTINEL-2-MSI-MultiRes-UTM30S",
            "SENTINEL-2-MSI-MultiRes-UTM31N",
            "SENTINEL-2-MSI-MultiRes-UTM31S",
            "SENTINEL-2-MSI-MultiRes-UTM32N",
            "SENTINEL-2-MSI-MultiRes-UTM32S",
            "SENTINEL-2-MSI-MultiRes-UTM33N",
            "SENTINEL-2-MSI-MultiRes-UTM33S",
            "SENTINEL-2-MSI-MultiRes-UTM34N",
            "SENTINEL-2-MSI-MultiRes-UTM34S",
            "SENTINEL-2-MSI-MultiRes-UTM35N",
            "SENTINEL-2-MSI-MultiRes-UTM35S",
            "SENTINEL-2-MSI-MultiRes-UTM36N",
            "SENTINEL-2-MSI-MultiRes-UTM36S",
            "SENTINEL-2-MSI-MultiRes-UTM37N",
            "SENTINEL-2-MSI-MultiRes-UTM37S",
            "SENTINEL-2-MSI-MultiRes-UTM38N",
            "SENTINEL-2-MSI-MultiRes-UTM38S",
            "SENTINEL-2-MSI-MultiRes-UTM39N",
            "SENTINEL-2-MSI-MultiRes-UTM39S",
            "SENTINEL-2-MSI-MultiRes-UTM40N",
            "SENTINEL-2-MSI-MultiRes-UTM40S",
            "SENTINEL-2-MSI-MultiRes-UTM41N",
            "SENTINEL-2-MSI-MultiRes-UTM41S",
            "SENTINEL-2-MSI-MultiRes-UTM42N",
            "SENTINEL-2-MSI-MultiRes-UTM42S",
            "SENTINEL-2-MSI-MultiRes-UTM43N",
            "SENTINEL-2-MSI-MultiRes-UTM43S",
            "SENTINEL-2-MSI-MultiRes-UTM44N",
            "SENTINEL-2-MSI-MultiRes-UTM44S",
            "SENTINEL-2-MSI-MultiRes-UTM45N",
            "SENTINEL-2-MSI-MultiRes-UTM45S",
            "SENTINEL-2-MSI-MultiRes-UTM46N",
            "SENTINEL-2-MSI-MultiRes-UTM46S",
            "SENTINEL-2-MSI-MultiRes-UTM47N",
            "SENTINEL-2-MSI-MultiRes-UTM47S",
            "SENTINEL-2-MSI-MultiRes-UTM48N",
            "SENTINEL-2-MSI-MultiRes-UTM48S",
            "SENTINEL-2-MSI-MultiRes-UTM49N",
            "SENTINEL-2-MSI-MultiRes-UTM49S",
            "SENTINEL-2-MSI-MultiRes-UTM50N",
            "SENTINEL-2-MSI-MultiRes-UTM50S",
            "SENTINEL-2-MSI-MultiRes-UTM51N",
            "SENTINEL-2-MSI-MultiRes-UTM51S",
            "SENTINEL-2-MSI-MultiRes-UTM52N",
            "SENTINEL-2-MSI-MultiRes-UTM52S",
            "SENTINEL-2-MSI-MultiRes-UTM53N",
            "SENTINEL-2-MSI-MultiRes-UTM53S",
            "SENTINEL-2-MSI-MultiRes-UTM54N",
            "SENTINEL-2-MSI-MultiRes-UTM54S",
            "SENTINEL-2-MSI-MultiRes-UTM55N",
            "SENTINEL-2-MSI-MultiRes-UTM55S",
            "SENTINEL-2-MSI-MultiRes-UTM56N",
            "SENTINEL-2-MSI-MultiRes-UTM56S",
            "SENTINEL-2-MSI-MultiRes-UTM57N",
            "SENTINEL-2-MSI-MultiRes-UTM57S",
            "SENTINEL-2-MSI-MultiRes-UTM58N",
            "SENTINEL-2-MSI-MultiRes-UTM58S",
            "SENTINEL-2-MSI-MultiRes-UTM59N",
            "SENTINEL-2-MSI-MultiRes-UTM59S",
            "SENTINEL-2-MSI-MultiRes-UTM60N",
            "SENTINEL-2-MSI-MultiRes-UTM60S",
            "SNAPHU",
            "Spacety",
            "SPOT-VGT",
            "SPOT4Take5",
            "SPOT5Take5",
            "SPOT6Dimap",
            "SPOTDimap",
            "SPOTView",
            "TerraSarX",
            "VIIRS-L1B",
            "VIIRS",
            "WorldView2GeoTIFF",
            "WorldViewGeoTIFF",
        ]

    @property
    def write_formats(self):
        return [
            "BEAM-DIMAP",
            "CSV",
            "ENVI",
            "Gamma for PyRate",
            "Gamma",
            "GDAL-BMP-WRITER",
            "GDAL-BT-WRITER",
            "GDAL-GS7BG-WRITER",
            "GDAL-GSBG-WRITER",
            "GDAL-GTiff-WRITER",
            "GDAL-GTX-WRITER",
            "GDAL-HFA-WRITER",
            "GDAL-ILWIS-WRITER",
            "GDAL-KRO-WRITER",
            "GDAL-MFF-WRITER",
            "GDAL-NITF-WRITER",
            "GDAL-PCIDSK-WRITER",
            "GDAL-PNM-WRITER",
            "GDAL-RMF-WRITER",
            "GDAL-RST-WRITER",
            "GDAL-SAGA-WRITER",
            "GDAL-SGI-WRITER",
            "Generic Binary BSQ",
            "GeoTIFF-BigTIFF",
            "GeoTIFF",
            "GeoTIFF+XML",
            "HDF5",
            "JP2",
            "JPEG2000",
            "NetCDF4-BEAM",
            "NetCDF4-CF",
            "PolSARPro",
            "Snaphu",
        ]



import attr
import lxml.etree as etree


@attr.s
class AggregatorSum(object):
    type = attr.ib(init=False, default="SUM")
    varName = attr.ib()
    targetName = attr.ib()

    def __str__(self):
        self.type = "SUM"
        return self.__repr__()

    def __repr__(self):
        self.type = "SUM"
        return "AggregatorSum({})".format(
            ", ".join(["{}='{}'".format(key, value) for key, value in self.to_dict().items()])
        )

    def to_dict(self):
        self.type = "SUM"
        return attr.asdict(self)

    def to_xml(self):
        self.type = "SUM"
        root = etree.Element("aggregator")

        for key, value in self.to_dict().items():
            elem = etree.SubElement(root, key)
            elem.text = str(value)

        return root



import logging
import sys

logger = logging.getLogger(__package__)


def setup_logger(level: int = logging.INFO) -> None:
    global logger
    logger.setLevel(level)
    stdout_handler = logging.StreamHandler(sys.stdout)
    logger.addHandler(stdout_handler)



import os
from base64 import b64encode
from urllib.parse import urlparse
from xml.sax.saxutils import unescape

import lxml.etree as etree
import requests
import yaml

from .binning import Aggregators, BinningVariables
from .binning.output_bands import BinningOutputBands
from .target_band_descriptors import TargetBandDescriptors


class Graph:
    """SNAP Graph class

    This class provides the methods to create, view and run a SNAP Graph

    Attributes:
        None.
    """

    def __init__(self, wdir=".", root=None):
        if root is None:
            self.root = etree.Element("graph")
            version = etree.SubElement(self.root, "version")
            version.text = "1.0"
        else:
            self.root = root

        self.pid = None
        self.p = None
        self.wdir = wdir

    def __str__(self):
        return "working dir: {}\n\n{}".format(
            self.wdir,
            etree.tostring(self.root, pretty_print=True).decode("utf-8"),
        ).replace("\\n", "\n")

    def __repr__(self):
        return "Graph(wdir='{}')".format(self.wdir)

    @classmethod
    def from_text(cls, text):
        root = etree.fromstring(text)
        return cls(root=root)

    @classmethod
    def from_uri(cls, uri):
        parsed = urlparse(uri)
        if parsed.scheme.startswith("http"):
            r = requests.get(uri)
            root = etree.fromstring(r.content).xpath("/graph")[0]
        else:
            root = etree.parse(uri).getroot()

        return cls(root=root)

    @staticmethod
    def list_operators():
        """This function provides a Python dictionary with all SNAP operators.

        Args:
            None.

        Returns
            Python dictionary with all SNAP operators.

        Raises:
            None.
        """
        filename = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data/operators.yaml")
        with open(filename) as f:
            operators = yaml.safe_load(f)

        return list(operators.keys())

    @staticmethod
    def describe_operators():
        """This function provides a Python dictionary with all SNAP operators.

        Args:
            None.

        Returns
            Python dictionary with all SNAP operators.

        Raises:
            None.
        """
        with open("data/operators.yaml") as f:
            operators = yaml.safe_load(f)

        descriptions = {key: value["desc"] for key, value in operators.items()}

        for op, desc in descriptions.items():
            print(f"{op} - {desc}")

        return descriptions

    def b64encode(self):
        graph = etree.tostring(self.root, pretty_print=True)
        return b64encode(graph).decode()

    def nice_view(self):
        try:
            import IPython
            from IPython.display import HTML
            from pygments import highlight
            from pygments.formatters import HtmlFormatter
            from pygments.lexers import XmlLexer

            def display_xml_nice(xml):
                formatter = HtmlFormatter()
                IPython.display.display(
                    HTML(
                        '<style type="text/css">{}</style>    {}'.format(
                            formatter.get_style_defs(".highlight"),
                            highlight(xml, XmlLexer(), formatter),
                        )
                    )
                )

            display_xml_nice(etree.tostring(self.root, pretty_print=True))

        except ModuleNotFoundError:
            print(
                etree.tostring(self.root, pretty_print=True).decode("utf-8").replace("\\n", "\n")
            )

    def view(self):
        """This method prints SNAP Graph

        Args:
            None.

        Returns
            None.

        Raises:
            None.
        """
        print(unescape(etree.tostring(self.root, pretty_print=True).decode("utf-8")))

    def add_node(self, operator, node_id, source=None):
        """This method adds or overwrites a node to the SNAP Graph

        Args:
            operator: SNAP operator
            node_id: node identifier
            source: string or list of sources (previous node identifiers in the SNAP Graph)

        Returns
            None.

        Raises:
            None.
        """
        xpath_expr = '/graph/node[@id="%s"]' % node_id

        if len(self.root.xpath(xpath_expr)) != 0:
            node_elem = self.root.xpath(xpath_expr)[0]
            operator_elem = self.root.xpath(xpath_expr + "/operator")[0]
            sources_elem = self.root.xpath(xpath_expr + "/sources")[0]
            parameters_elem = self.root.xpath(xpath_expr + "/parameters")

            for param in [
                name
                for name in dir(operator)
                if not name.startswith("__")
                and not name.endswith("__")
                and name != "_params"
                and name != "operator"
                and type(getattr(operator, name)).__name__
                in [
                    "str",
                    "NoneType",
                    "TargetBandDescriptors",
                    "Aggregators",
                    "BinningOutputBands",
                    "BinningVariables",
                ]
            ]:
                if param in [
                    "targetBandDescriptors",
                    "aggregatorConfigs",
                    "variableConfigs",
                    "bandConfigurations",
                    "postProcessorConfig",
                    "productCustomizerConfig",
                ]:
                    if param in [
                        "bandConfigurations",
                        "variableConfigs",
                        "postProcessorConfig",
                        "productCustomizerConfig",
                    ] and not getattr(operator, param):
                        continue

                    if (
                        isinstance(getattr(operator, param), TargetBandDescriptors)
                        or isinstance(getattr(operator, param), Aggregators)
                        or isinstance(getattr(operator, param), BinningOutputBands)
                        or isinstance(getattr(operator, param), BinningVariables)
                    ):
                        parameters_elem.append(getattr(operator, param).to_xml())

                    elif isinstance(getattr(operator, param), str):
                        parameters_elem.append(etree.fromstring(getattr(operator, param)))
                    else:
                        raise ValueError()

                else:
                    try:
                        p_elem = self.root.xpath(xpath_expr + "/parameters/%s" % param)[0]

                        if getattr(operator, param) is not None:
                            if getattr(operator, param)[0] != "<":
                                p_elem.text = getattr(operator, param)
                            else:
                                p_elem.text.append(etree.fromstring(getattr(operator, param)))
                    except IndexError:
                        pass

        else:
            node_elem = etree.SubElement(self.root, "node")
            operator_elem = etree.SubElement(node_elem, "operator")
            sources_elem = etree.SubElement(node_elem, "sources")

            if isinstance(source, list):
                for index, s in enumerate(source):
                    if index == 0:
                        source_product_elem = etree.SubElement(sources_elem, "sourceProduct")

                    else:
                        source_product_elem = etree.SubElement(
                            sources_elem, "sourceProduct.%s" % str(index)
                        )

                    source_product_elem.attrib["refid"] = s

            elif isinstance(source, dict):
                for key, value in source.iteritems():
                    source_product_elem = etree.SubElement(sources_elem, key)
                    source_product_elem.text = value

            elif source is not None:
                source_product_elem = etree.SubElement(sources_elem, "sourceProduct")
                source_product_elem.attrib["refid"] = source

            parameters_elem = etree.SubElement(node_elem, "parameters")
            parameters_elem.attrib["class"] = "com.bc.ceres.binding.dom.XppDomElement"

            for param in [
                name
                for name in dir(operator)
                if not name.startswith("__")
                and not name.endswith("__")
                and name != "_params"
                and name != "operator"
                and type(getattr(operator, name)).__name__
                in [
                    "str",
                    "NoneType",
                    "TargetBandDescriptors",
                    "Aggregators",
                    "BinningOutputBands",
                    "BinningVariables",
                ]
            ]:
                if param in [
                    "targetBandDescriptors",
                    "aggregatorConfigs",
                    "variableConfigs",
                    "bandConfigurations",
                    "postProcessorConfig",
                    "productCustomizerConfig",
                ]:
                    print(param, getattr(operator, param))
                    if param in [
                        "bandConfigurations",
                        "variableConfigs",
                        "postProcessorConfig",
                        "productCustomizerConfig",
                    ] and not getattr(operator, param):
                        continue

                    if (
                        isinstance(getattr(operator, param), TargetBandDescriptors)
                        or isinstance(getattr(operator, param), Aggregators)
                        or isinstance(getattr(operator, param), BinningOutputBands)
                        or isinstance(getattr(operator, param), BinningVariables)
                    ):
                        parameters_elem.append(getattr(operator, param).to_xml())

                    elif isinstance(getattr(operator, param), str):
                        parameters_elem.append(etree.fromstring(getattr(operator, param)))
                    else:
                        raise ValueError()

                else:
                    parameter_elem = etree.SubElement(parameters_elem, param)

                    if getattr(operator, param) is not None:
                        if getattr(operator, param)[0] != "<":
                            parameter_elem.text = getattr(operator, param)
                        else:
                            parameter_elem.append(etree.fromstring(getattr(operator, param)))

        node_elem.attrib["id"] = node_id

        operator_elem.text = operator.operator

    def save_graph(self, filename):
        """This method saves the SNAP Graph

        Args:
            filename: XML filename with '.xml' extension

        Returns
            None.

        Raises:
            None.
        """

        with open(filename, "w") as file:
            file.write('<?xml version="1.0" encoding="UTF-8"?>\n')
            file.write(unescape(etree.tostring(self.root, pretty_print=True).decode()))



import attr
import lxml.etree as etree


@attr.s
class AggregatorPercentile(object):
    type = attr.ib(init=False, default="PERCENTILE")
    varName = attr.ib()
    targetName = attr.ib()
    percentage = attr.ib(default=90)

    @percentage.validator
    def _check_type(self, attribute, value):
        if not (isinstance(value, int)):
            raise ValueError("percentage is an int")

        if int(value) < 0 or int(value > 100):
            raise ValueError("percentage value is in range [0.0, 100.0]")

    def __str__(self):
        self.type = "PERCENTILE"
        return self.__repr__()

    def __repr__(self):
        self.type = "PERCENTILE"
        return "AggregatorPercentile({})".format(
            ", ".join(["{}='{}'".format(key, value) for key, value in self.to_dict().items()])
        )

    def to_dict(self):
        self.type = "PERCENTILE"

        return attr.asdict(self)

    def to_xml(self):
        self.type = "PERCENTILE"

        root = etree.Element("aggregator")

        for key, value in self.to_dict().items():
            elem = etree.SubElement(root, key)
            elem.text = str(value)

        return root



# pytype: skip-file

"""
This module contains a utility to extract and save to a local json file all the
SNAP operators metadata.
It requires a full Snapista installation through conda or mamba and a local SNAP
instance, which comes with Snapista.
"""

import json
from typing import Dict

from snapista import Graph
from snappy import GPF, jpy

from .snap_types import OperatorType, ParamType


def get_formats(method):
    manager = jpy.get_type("org.esa.snap.core.dataio.ProductIOPlugInManager")
    if method == "Read":
        plugins = manager.getInstance().getAllReaderPlugIns()
    elif method == "Write":
        plugins = manager.getInstance().getAllWriterPlugIns()
    else:
        raise ValueError
    formats = []
    while plugins.hasNext():
        plugin = plugins.next()
        formats.append(plugin.getFormatNames()[0])
    return formats


def get_snap_operators() -> Dict[str, OperatorType]:
    operators = Graph.list_operators()
    d = {o: {} for o in operators}
    for o in operators:
        spi = GPF.getDefaultInstance().getOperatorSpiRegistry().getOperatorSpi(o)
        d[o]["alias"] = spi.getOperatorDescriptor().getAlias()
        d[o]["description"] = spi.getOperatorDescriptor().getDescription()
        d[o]["authors"] = spi.getOperatorDescriptor().getAuthors()
        d[o]["name"] = spi.getOperatorDescriptor().getName()
        d[o]["version"] = spi.getOperatorDescriptor().getVersion()
        d[o]["params"] = {}
        params = spi.getOperatorDescriptor().getParameterDescriptors()
        for param in params:
            d[o]["params"][param.getName()] = {}
            d[o]["params"][param.getName()]["name"] = param.getName()
            d[o]["params"][param.getName()]["description"] = param.getDescription()
            d[o]["params"][param.getName()]["default_values"] = param.getDefaultValue()
            if o == "Write" and param.getName() == "formatName":
                d[o]["params"][param.getName()]["values_set"] = get_formats("Write")
            elif o == "Read" and param.getName() == "formatName":
                d[o]["params"][param.getName()]["values_set"] = get_formats("Read")
            else:
                d[o]["params"][param.getName()]["values_set"] = list(param.getValueSet())
    return d


def serialize(operators: Dict[str, OperatorType], filename: str):
    with open(filename, "w") as f:
        json.dump(operators, f)


def deserialize(filename: str) -> Dict[str, OperatorType]:
    with open(filename) as f:
        operators = json.load(f)

    operators = {k: OperatorType(**v) for k, v in operators.items()}
    for k, v in operators.items():
        v["params"] = [ParamType(**p) for p in v["params"]]

    return operators



from .aggregators import Aggregators  # noqa: F401
from .binning_band import BinningBand  # noqa: F401
from .binning_variable import BinningVariable  # noqa: F401
from .binning_variables import BinningVariables  # noqa: F401
from .output_bands import BinningOutputBands  # noqa: F401



class EOCanvasError(Exception):
    """Base class for EO Canvas exceptions."""


class CredentialsError(EOCanvasError):
    """Generic error regarding the credentials."""


class MalformedCredentialsError(CredentialsError):
    """Malformed credentials error."""


class CorruptedCredentialsError(CredentialsError):
    """Credentials file is bad YAML."""


class CredentialsNotFoundError(CredentialsError):
    """Missing credentials file."""


class HTTPError(EOCanvasError):
    """Exception on low level HTTP errors."""


class QuotaExceededError(EOCanvasError):
    """Exception on user quota exceeded."""


class MalformedSnapError(EOCanvasError):
    """Exception on malformed SNAP operators file."""


class SnapOperatorNotFound(EOCanvasError):
    """Exception on non-existent SNAP operator."""


class JobFailed(EOCanvasError):
    """Exception on a job that returns failed status."""


class InvalidChainError(EOCanvasError):
    """Exception on invalid YAML file for a Data Tailor Chain."""


class NotDownloadableError(EOCanvasError):
    """Exception on results that are not downloadable, coming from an external source"""


class UnknownResultTypeError(EOCanvasError):
    """Results of unknown rel type."""



import attr
import lxml.etree as etree


@attr.s
class AggregatorAvg(object):
    type = attr.ib(init=False, default="AVG")
    varName = attr.ib()
    targetName = attr.ib()
    weightCoeff = attr.ib(default=0.0)
    outputCounts = attr.ib(default="true")
    outputSums = attr.ib(default="true")

    @outputCounts.validator
    def _check_counts_type(self, attribute, value):
        if value not in ["true", "false"]:
            raise ValueError("output_counts value is either 'true' or 'false'")

    @outputSums.validator
    def _check_sums_type(self, attribute, value):
        if value not in ["true", "false"]:
            raise ValueError("output_sums value is either 'true' or 'false'")

    def __str__(self):
        self.type = "AVG"
        return self.__repr__()

    def __repr__(self):
        self.type = "AVG"
        return "AggregatorAvg({})".format(
            ", ".join(["{}='{}'".format(key, value) for key, value in self.to_dict().items()])
        )

    def to_dict(self):
        self.type = "AVG"
        return attr.asdict(self)

    def to_xml(self):
        self.type = "AVG"
        root = etree.Element("aggregator")

        for key, value in self.to_dict().items():
            elem = etree.SubElement(root, key)
            elem.text = str(value)

        return root



from __future__ import annotations

import re
import time
from urllib.parse import urljoin

import requests
import yaml
from requests.auth import AuthBase

from .config import get_credentials_filepath
from .exceptions import (
    CorruptedCredentialsError,
    CredentialsNotFoundError,
    MalformedCredentialsError,
)
from .http import post
from .logging import logger


class Credentials:
    """Provides basic credentials functionalities, such as save on or load from file."""

    def __init__(self, username: str, password: str) -> None:
        umatch = re.match(r"[^\s]+$", username)
        pmatch = re.match(r"[^\s]+$", password)
        if umatch is None or pmatch is None:
            raise MalformedCredentialsError(
                "Username and/or password malformed. Please review them."
            )
        self.username = username
        self.password = password

    def save(self) -> None:
        credentials_path = get_credentials_filepath(True)
        try:
            credentials = {
                "user": self.username,
                "password": self.password,
            }
            with credentials_path.open(mode="w") as f:
                yaml.dump(credentials, f, default_flow_style=False)

            logger.info(f"Credentials are written to file {credentials_path}")
        except OSError:
            logger.error(
                "Credentials could not be written to {credentials_path}."
                "Please review your configuration."
            )

    @classmethod
    def load(cls) -> Credentials:
        credentials_path = get_credentials_filepath()
        try:
            with open(credentials_path) as f:
                credentials = yaml.safe_load(f)
                return cls(credentials["user"], credentials["password"])
        except (KeyError, TypeError):
            raise CorruptedCredentialsError(
                f"Credentials could not be read from {credentials_path}. Please save them again."
            )
        except FileNotFoundError:
            raise CredentialsNotFoundError(
                f"Missing credentials file at {credentials_path}. Please save them again."
            )


class OAuthToken:
    """OAuth token handler."""

    def __init__(self, url: str, credentials: Credentials, verify_ssl=True):
        self.url = url
        self.credentials = credentials
        self.verify_ssl = verify_ssl
        self._access_token = None
        self._refresh_token = None
        self._expiration_time = None

    def __str__(self) -> str:
        return self.access_token

    @property
    def is_expired(self) -> bool:
        now = int(time.time())
        return self._expiration_time is None or now > self._expiration_time

    @property
    def access_token(self) -> str:
        """The access token to the API.

        :return: A valid OAuth Access Token.
        :rtype: str
        """
        if self.is_expired:
            logger.debug("Token expired or invalid. Requesting a new one.")
            self._set_token()

        return self._access_token

    def invalidate(self) -> None:
        self._expiration_time = None

    def _set_token(self) -> None:
        """Requests and set a new access token using the configured credentials."""

        def get_token():
            return post(
                urljoin(self.url, "gettoken"),
                json={
                    "username": self.credentials.username,
                    "password": self.credentials.password,
                },
                verify=self.verify_ssl,
            )

        def refresh_token():
            return post(
                urljoin(self.url, "refreshtoken"),
                data={"refresh_token": self._refresh_token},
                verify=self.verify_ssl,
            )

        if self._refresh_token is not None:
            response = refresh_token()
            if response.status_code in (
                requests.codes.forbidden,
                requests.codes.bad_request,
            ):
                response = get_token()
        else:
            response = get_token()

        response.raise_for_status()
        data = response.json()
        self._access_token = data["access_token"]
        self._refresh_token = data["refresh_token"]
        self._expiration_time = int(time.time()) + data["expires_in"]


class HTTPOAuth2(AuthBase):
    """HTTP authentication to be used with requests. Set an bearer authorization header."""

    def __init__(self, token: OAuthToken):
        self.token = token

    def __call__(self, request: requests.PreparedRequest) -> requests.PreparedRequest:
        request.headers.update(self.header)
        return request

    @property
    def header(self) -> dict:
        return {"authorization": f"Bearer {self.token}"}



import json
from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, List

from .api import Config, Input, Process
from .datatailor.chain import Chain
from .snap.graph import Graph


@dataclass
class _SnapParams:
    snap_graph: Graph
    eo_input: List[Input] = field(default_factory=list)
    eo_config: List[Config] = field(default_factory=list)


@dataclass
class SnapProcess(Process, _SnapParams):
    def __post_init__(self):
        super().__post_init__()
        if self.process_id is None:
            self.process_id = "snap-function"

        if not isinstance(self.eo_config, list):
            self.eo_config = [self.eo_config]

        if not isinstance(self.eo_input, list):
            self.eo_input = [self.eo_input]

    def prepare_inputs(self) -> Dict:
        inputs = super().prepare_inputs()
        inputs["inputs"] = {
            "snap_graph": self.snap_graph.b64encode(),
            "eo_input": json.dumps([i.asdict() for i in self.eo_input]),
        }
        if self.eo_config:
            config_list = [c.asdict() for c in self.eo_config]
            inputs["inputs"]["eo_config"] = json.dumps(config_list)

        return inputs


@dataclass
class _DataTailorParams:
    epct_chain: Chain
    epct_input: List[Input] = field(default_factory=list)
    epct_config: List[Config] = field(default_factory=list)


@dataclass
class DataTailorProcess(Process, _DataTailorParams):
    def __post_init__(self):
        super().__post_init__()
        if self.process_id is None:
            self.process_id = "dataTailor"

        if not isinstance(self.epct_config, list):
            self.epct_config = [self.epct_config]

        if not isinstance(self.epct_input, list):
            self.epct_input = [self.epct_input]

    def prepare_inputs(self) -> Dict:
        inputs = super().prepare_inputs()
        inputs["inputs"] = {
            "epct_chain": self.epct_chain.b64encode(),
            "epct_input": json.dumps([i.asdict() for i in self.epct_input]),
        }
        if self.epct_config:
            config_list = [c.asdict() for c in self.epct_config]
            inputs["inputs"]["epct_config"] = json.dumps(config_list)

        return inputs


@dataclass
class ShearWaterParams:
    area: str
    start_day: str
    end_day: str

    valid_areas: set = field(default_factory=lambda: {"Sindian"}, init=False, repr=False)


@dataclass
class ShearWaterProcess(Process, ShearWaterParams):
    def __post_init__(self):
        super().__post_init__()
        if self.process_id is None:
            self.process_id = "shearwater-demo"

        if self.area not in self.valid_areas:
            raise ValueError(f"Invalid area '{self.area}'. Must be one of {self.valid_areas}.")

        for date_field_name in ["start_day", "end_day"]:
            date_value = getattr(self, date_field_name)
            try:
                datetime.strptime(date_value, "%Y-%m-%d")
            except ValueError:
                raise ValueError(f"{date_field_name} '{date_value}' is not in YYYY-MM-DD format.")

    def prepare_inputs(self) -> Dict:
        inputs = super().prepare_inputs()
        inputs["inputs"] = {
            "area": self.area,
            "startDay": self.start_day,
            "endDay": self.end_day,
        }
